//=============================================================================
// Copyright (c) 2016-present Allan CORNET (Nelson)
//=============================================================================
// This file is part of the Nelson.
//=============================================================================
// LICENCE_BLOCK_BEGIN
// SPDX-License-Identifier: LGPL-3.0-or-later
// LICENCE_BLOCK_END
//=============================================================================
#include <boost/chrono/chrono.hpp>
#include <boost/thread/thread.hpp>
#include "pauseBuiltin.hpp"
#include "Error.hpp"
#include "OverloadFunction.hpp"
#include "OverloadRequired.hpp"
#include "ProcessEventsDynamicFunction.hpp"
#include "NelsonConfiguration.hpp"
//=============================================================================
using namespace Nelson;
//=============================================================================
static bool pauseOn = true;
//=============================================================================
ArrayOfVector
Nelson::CoreGateway::pauseBuiltin(Evaluator* eval, int nLhs, const ArrayOfVector& argIn)
{
    ArrayOfVector retval;
    nargoutcheck(nLhs, 0, 1);
    nargincheck(argIn, 0, 1);
    if (argIn.size() == 0) {
        if (nLhs == 1) {
            if (pauseOn) {
                retval << ArrayOf::characterArrayConstructor(L"on");
            } else {
                retval << ArrayOf::characterArrayConstructor(L"off");
            }
            return retval;
        }
        Interface* io = eval->getInterface();
        if (io != nullptr) {
            io->getInput(L"");
        }
    } else {
        ArrayOf param1 = argIn[0];
        if (param1.isRowVectorCharacterArray()) {
            std::wstring arg1Value = param1.getContentAsWideString();
            if (arg1Value == L"on" || arg1Value == L"off" || arg1Value == L"query") {
                bool previousValue = pauseOn;
                if (arg1Value == L"on") {
                    pauseOn = true;
                    if (nLhs == 1) {
                        if (previousValue) {
                            retval << ArrayOf::characterArrayConstructor(L"on");
                        } else {
                            retval << ArrayOf::characterArrayConstructor(L"off");
                        }
                        return retval;
                    }
                } else if (arg1Value == L"off") {
                    pauseOn = false;
                    if (nLhs == 1) {
                        if (previousValue) {
                            retval << ArrayOf::characterArrayConstructor(L"on");
                        } else {
                            retval << ArrayOf::characterArrayConstructor(L"off");
                        }
                        return retval;
                    }
                } else {
                    if (pauseOn) {
                        retval << ArrayOf::characterArrayConstructor(L"on");
                    } else {
                        retval << ArrayOf::characterArrayConstructor(L"off");
                    }
                    return retval;
                }
            } else {
                Error(ERROR_WRONG_ARGUMENT_1_VALUE);
            }
        } else if (param1.isNumeric()) {
            if (nLhs == 1) {
                if (pauseOn) {
                    retval << ArrayOf::characterArrayConstructor(L"on");
                } else {
                    retval << ArrayOf::characterArrayConstructor(L"off");
                }
                return retval;
            }
            double val = param1.getContentAsDoubleScalar();
            if (!pauseOn) {
                return retval;
            }
            if (std::isinf(val)) {
                while (!NelsonConfiguration::getInstance()->getInterruptPending()) {
                    boost::this_thread::sleep_for(boost::chrono::milliseconds(uint64(10)));
                    if (eval->haveEventsLoop()) {
                        ProcessEventsDynamicFunctionWithoutWait();
                    }
                }
            } else if (std::isnan(val)) {
                // DO NOTHING
            } else {
                boost::chrono::nanoseconds begin_time
                    = boost::chrono::high_resolution_clock::now().time_since_epoch();
                bool bContinue = true;
                do {
                    boost::this_thread::sleep_for(boost::chrono::nanoseconds(uint64(10)));
                    boost::chrono::nanoseconds current_time
                        = boost::chrono::high_resolution_clock::now().time_since_epoch();
                    boost::chrono::nanoseconds difftime = (current_time - begin_time);
                    bContinue = !(difftime.count() > int64(val * 1e9));
                    if (eval->haveEventsLoop()) {
                        ProcessEventsDynamicFunctionWithoutWait();
                    }
                } while (!NelsonConfiguration::getInstance()->getInterruptPending()
                    && (static_cast<int>(bContinue) == true));
            }

        } else {
            bool bSuccess = false;
            retval = OverloadFunction(eval, nLhs, argIn, "pause", bSuccess);
            if (!bSuccess) {
                OverloadRequired(eval, argIn, Overload::OverloadClass::FUNCTION, "pause");
            }
        }
    }
    return retval;
}
//=============================================================================
